/* Supports automatically extended slot arrays of arbitrary type */


#ifndef _SLOTS_H
#define _SLOTS_H 1


/*********
 * Types *
 *********/


/* A weakly typed slot array. The 'array' member points at the actual
   slot elements, which do not need to derive from anything. However,
   if you want to use the macros, the actual type needs to include a
   declaration SLOTTABLE */

typedef struct slots {
    void *array;
    int size, count;
} SLOTS_T;


/* Strongly typed slot array. Can be used anywhere instead of the weakly
   typed version. */

#define TSLOTS_T(type)	    struct { type *array; int size, count; }


/* Every structure that is to be stored in slots must have this declaration 
   somewhere. This member only accessed by macros, hence it does not need to
   be at a fixed place. */

#define SLOTTABLE	    int inuse


/**********
 * Macros *
 **********/


/* Return a typecast pointer to a slot, or 0 if the slot doesn't exist,
   i.e. is beyond the array (the inuse mark is irrelevant here). */

#define SLOTP(slots, index, type)				\
    ((type *)((unsigned)(index) < (slots).count ?		\
              ((char *)(slots).array) + (slots).size * (index) : 0)) 

/* Return a typed pointer to a strongly typed slot */

#define TSLOTP(slots, index)					\
	((unsigned)(index) < (slots).count ? (slots).array + (index) : 0)


/* Return a member (of a member...) of a weakly or strongly typed slotted
   structure, or a default (l)value if the slot doesn't exist or is marked
   unused. */

#define SLOTMB(slots, index, member, default, type)		\
    ((unsigned)(index) < (slots).count &&			\
     ((type *)((slots).array))[index].inuse ?			\
     ((type *)((slots).array))[index].member : (default))

#define TSLOTMB(slots, index, member, default)			\
    ((unsigned)(index) < (slots).count &&			\
     (slots).array[index].inuse ?				\
     (slots).array[index].member : (default))


/* Mark a slot pointed to by typed pointer sp used/unused */

#define SLOT_INUSE(sp)		do { if ((sp)) (sp)->inuse = 1; } while(0)
#define SLOT_RELEASE(sp)	do { if ((sp)) (sp)->inuse = 0; } while(0)

/* Test whether a slot at sp is in use */

#define SLOT_ISINUSE(sp)	((sp) && (sp)->inuse)


/* Return a slot number based on a slot pointer, or -1 if the pointer does
   not point inside the slot array. Does not care about in use status; that's 
   one level higher. */

#define SLOTNR(slots, sp)						\
	((char *)(sp) >= (char *)(slots).array &&			\
	 (char *)(sp) < (char *)(slots).array + (slots).count * (slots).size ? \
	 ((char *)(sp) - (char *)(slots).array) / (slots).size : -1)


/* Iterate over all slots; idxvar holds slot number, ptrvar slot pointer */

#define SLOTS_FOREACHIDX(slots, idxvar)				\
    for((idxvar) = 0; (idxvar) < (slots).count; (idxvar)++)

#define SLOTS_FOREACH(slots, ptrvar, idxvar)			\
    for((idxvar) = 0, (ptrvar) = (typeof(ptrvar))(slots).array;	\
	(idxvar) < (slots).count;				\
	(idxvar)++, (ptrvar) = (typeof(ptrvar))((char *)(ptrvar)+(slots).size))

#define TSLOTS_FOREACH(slots, ptrvar, idxvar)			\
    for((idxvar) = 0, (ptrvar) = (slots).array;			\
	(idxvar) < (slots).count;				\
	(idxvar)++, (ptrvar)++)


/* Iterate over all occupied slots */

#define SLOTS_FOREACHINUSE(slots, ptrvar, idxvar)		\
    SLOTS_FOREACH((slots), (ptrvar), (idxvar))			\
	if ((ptrvar)->inuse)

#define TSLOTS_FOREACHINUSE(slots, ptrvar, idxvar)		\
    TSLOTS_FOREACH((slots), (ptrvar), (idxvar))			\
	if ((ptrvar)->inuse)


/* This calculates cnt + trunc(cnt * 0.3125) (ie. 5/16). The factor is chosen
   so that after two reallocations, a memory hole exists that is big enough
   for the third to fit in. We avoid creating an exponential memory hole this
   way. */

#define SLOTS_PCTINCREASE(cnt)					\
    ((cnt) + ((((cnt) << 2) + (cnt)) >> 4))


/**************
 * PROTOTYPES *
 **************/


/* Constructors and destructors. Note: the initial slot count given here is
   just a hint; there's no guarantee that you actually have the slots after
   calling this. In other words, it doesn't discharge you from having to call
   slots_brk! */

int slots_init(SLOTS_T *slots, int slotsize, int initcount);
void slots_done(SLOTS_T *slots);

#define tslots_init(slots, initcount)				\
	slots_init((SLOTS_T *)(slots), sizeof((slots)->array[0]), (initcount))

#define tslots_done(slots)					\
	slots_done((SLOTS_T *)(slots))


/* Tries to make sure sure that the element at 'index' can be accessed, ie. 
   that the slot array has at least index + 1 slots available. Returns the
   number of actual slots available.  Note: this is definitely *not*
   thread-safe; if you must do multithreading, guard a slot array structure
   with a lock during BRK and DEL. */

int slots_brk(SLOTS_T *slots, int index);

#define tslots_brk(slots, index)				\
	slots_brk((SLOTS_T *)(slots), (index))


#endif	    /* _SLOTS_H */


/*
 * vim:softtabstop=4:sw=4
 */
