/* Evproc defines an interface between an event driven subprocess and a
   select()/poll()/WaitForMultipleObjects() loop that multiplexes those events.

   Anything that has an evproc interface can be aggregated into an evprocset,
   which itself also implements the evproc interface, allowing event processes
   to be nested recursively.

   The handleset provided by each evproc can be waited on by handleset_wait,
   which returns a handleset specifying the set of signalled handles after
   waiting.

   After waiting, you can ask an evproc to try and perform any operation it
   wants to, based on the set of signalled handles and the current time; 
   including the evprocset, which will in turn ask all its evproc members. */


/* INCLUDES & DEFINES */

#ifndef _EVPROC_H
#define _EVPROC_H 1

#include <evblib/evproc/handleset.h>

#include <evblib/db/list.h>


/* TYPES */

/* The base structure for each event process; basically a function table and a 
   private member for simple processes that don't need to derive their own */

typedef struct evproc 
{
	/* The function pointers */

#define F(type, name, args...)	type (*name)( args );
#include <evblib/evproc/evproc.inc>
#undef F

	/* Local stuff */

	void *private;

} EVPROC_T;

/* Box type to allow evprocs to be managed by the evprocset */

typedef enum evprocbox_endstyle
{ 
    EP_REM_NOP, 
    EP_REM_CALLDONE, 
    EP_REM_CALLDONE_FREE 

} evprocbox_endstyle_t;

typedef struct evprocbox
{
    LISTABLE( struct evprocbox );
    EVPROC_T *ep;
    evprocbox_endstyle_t endstyle;

} EVPROCBOX_T;

typedef struct evprocset 
{
    EVPROC_T ep;			/* base class; we're an EVPROC_T too */
    HANDLESET_T cachedhs;		/* cached union of handle sets */
    EVPROCBOX_T *ephead, *eptail;	/* children */

} EVPROCSET_T;


/* PROTOTYPES */

/* Dummy evproc interface */

/* Provides null_ep_init, _ep_addhandles, _ep_handle, _ep_done prototypes; 
   useful if you don't need to define a particular function yourself. */

#define F(type, name, args...)	type null_##name( args ); 
#include <evblib/evproc/evproc.inc> 
#undef F

/* Evprocset's evproc interface; defines evprocset_ep_*() */

#define F(type, name, args...)	type evprocset_##name( args ); 
#include <evblib/evproc/evproc.inc> 
#undef F

/* Evprocset's own interface */

/* evprocset_add adds an evproc to the evprocset. Returns pointer to the box
 * created for it on success, 0 on failure. If nonzero is returned, guarantees
 * that the ep member of the box is nonzero as well. The endstyle argument 
 * is copied on the box and specifies how the evproc inside should be treated 
 * upon removal from the evprocset: 
 *   EP_REM_NOP		    - leave it alone
 *   EP_REM_CALLDONE	    - call ep->ep_done( ep )
 *   EP_REM_CALLDONE_FREE   - call ep->ep_done( ep ) and call free( ep )
 *
 * evprocset_rem removes an evprocbox from the evprocset, dealing with the
 * evproc inside according to the endstyle specified on the box.
 *
 * evprocset_update must be called to update evprocset's handle cache when any 
 * evproc in the set has changed the set added by its ep_addhandles. 
 *
 * evprocset_clear effectively calls evprocset_rem for all evprocs in the set */

void evprocset_init( EVPROCSET_T *eps);

EVPROCBOX_T *evprocset_add( EVPROCSET_T *eps, EVPROC_T *ep, 
			    evprocbox_endstyle_t endstyle );   

void evprocset_rem( EVPROCSET_T *eps, EVPROCBOX_T *epb );
void evprocset_update( EVPROCSET_T *eps );
void evprocset_clear( EVPROCSET_T *eps );	

void evprocset_done( EVPROCSET_T *eps );

#endif		/* EVPROC_H */

/*
 * vim:softtabstop=4:sw=4
 */

