/*
 * STR - Support for explicit length strings. These have the advantage of
 *       allowing to pass around substrings without copying, to use the length
 *       as a quick hash, and to contain the \0 character.
 *
 * Author:
 * Emile van Bergen, emile@evbergen.xs4all.nl
 *
 * Permission to redistribute an original or modified version of this program
 * in source, intermediate or object code form is hereby granted exclusively
 * under the terms of the GNU General Public License, version 2. Please see the
 * file COPYING for details, or refer to http://www.gnu.org/copyleft/gpl.html.
 *
 * History:
 * 2004/01/13 - EvB - Created
 * 2004/05/11 - EvB - Split off from misc
 * 2005/03/23 - EvB - Copied from OpenRADIUS
 * 2005/05/23 - EvB - Added FREESTR macro and str_dup, str_safedup functions
 */


char str_id[] = "STR - Copyright (C) 2004 Emile van Bergen.";


/*
 * INCLUDES & DEFINES
 */


#include <string.h>		/* For memchr */
#include <ctype.h>		/* For toupper, tolower */

#include <evblib/misc/misc.h>	/* For MIN */
#include <evblib/str/str.h>


/*
 * FUNCTIONS
 */


void *memrchr(void *s, int c, size_t n)
{
    void *ret;
    int t;

    t = 1;				/* catch n == 0 case */
    for(ret = s + n - 1; 
	ret >= s && (t = *(unsigned char *)ret - (unsigned char)c);
	ret--);

    if (t == 0) return ret;
    return 0;
}


size_t memspn(void *data, size_t len, char *set, int setlen)
{
    char *p = (char *)data;

    switch(setlen) {
	case 1:
	    while(len && *p == *set) len--, p++;
	    return p - (char *)data;
	case 2:
	    while(len && (*p == set[0] ||
			      *p == set[1])) len--, p++;
	    return p - (char *)data;
	case 3:
	    while(len && (*p == set[0] ||
			      *p == set[1] ||
			      *p == set[2])) len--, p++;
	    return p - (char *)data;
	case 4:
	    while(len && (*p == set[0] ||
			      *p == set[1] ||
			      *p == set[2] ||
			      *p == set[3])) len--, p++;
	    return p - (char *)data;
    }
    /* any set length */
    while(len && memchr(set, *p, setlen)) len--, p++;
    return p - (char *)data;
}


size_t memcspn(void *data, size_t len, char *set, int setlen)
{
    char *p = (char *)data;

    switch(setlen) {
	case 1:
	    /* could be optimized further using (built-in) memchr */
	    while(len && *p != *set) len--, p++;
	    return p - (char *)data;
	case 2:
	    while(len && (*p != set[0] &&
			      *p != set[1])) len--, p++;
	    return p - (char *)data;
	case 3:
	    while(len && (*p != set[0] &&
			      *p != set[1] &&
			      *p != set[2])) len--, p++;
	    return p - (char *)data;
	case 4:
	    while(len && (*p != set[0] &&
			      *p != set[1] &&
			      *p != set[2] &&
			      *p != set[3])) len--, p++;
	    return p - (char *)data;
    }
    /* any set length */
    while(len && memchr(set, *p, setlen) == 0) len--, p++;
    return p - (char *)data;
}

/* returns string before first delimiter in s (empty if s starts with one, full
   string if none present) keeps s at ending delimiter (or end of string) */

STR_T str_firstword(STR_T *s, char delim)
{
	STR_T ret;
	char *e;

	if (s->l <= 0) return *s;
	e = memchr(s->p, delim, s->l);
	if (!e) return *s;
	ret.l = e - s->p;
	ret.p = s->p;
	s->p += ret.l;
	s->l -= ret.l;
	return ret;
}

/* removes one leading delimiter from s, if any, returns the part of s until
   the next delimiter or the end of the string, and adjusts s to start just
   beyond the returned string. */

STR_T str_nextword(STR_T *s, char delim)
{
	if (s->l <= 0) return *s;
	if (*s->p == delim) s->p++, s->l--;
	return str_firstword(s, delim);
}

/* returns bare word in STR_T, adjusts s to position after returned string. */

STR_T str_getword(STR_T *s, char *delim, int delimcnt) 
{
   STR_T ret;
   int n;

   n = memspn(s->p, s->l, delim, delimcnt);
   s->p += n; s->l -= n;

   n = memcspn(s->p, s->l, delim, delimcnt);
   ret.p = s->p;
   ret.l = n;
   s->p += n; s->l -= n;

   return ret;
}


/* copy string, limit size to destination buffer  */

ssize_t cpstr(char *d, ssize_t dl, char *s, ssize_t sl)
{
    ssize_t ret; 

    ret = MIN(sl, dl);
    if (ret > 0) {
       memcpy(d, s, ret);
       return ret;
    }
    return 0;
}


/* copy and terminate string, limit terminated size to destination buffer */

ssize_t cptstr(char *d, ssize_t dl, char *s, ssize_t sl)
{
    ssize_t ret; 

    ret = MIN(sl, dl - 1);
    if (ret > 0) {
       memcpy(d, s, ret);
       d[ret] = 0;
       return ret;
    }
    *d = 0;
    return 0;
}


void str_toupper(char *p, ssize_t l)
{
    while(l-- > 0) { *p = toupper(*p); p++; }
}


void str_tolower(char *p, ssize_t l)
{
    while(l-- > 0) { *p = tolower(*p); p++; }
}


STR_T str_dup(char *p, ssize_t l)
{
    STR_T ret = {0, 0};

    if (l > 0) {
	ret.p = (char *)malloc(l);
	if (ret.p) {
	    ret.l = l;
	    memcpy(ret.p, p, l);
	}
    }
    return ret;
}


STR_T str_safedup(char *p, ssize_t l)
{
    STR_T ret = {0, 0};

    if (l > 0) {
	ret.p = (char *)safe_malloc(l);
	ret.l = l;
	memcpy(ret.p, p, l);
    }
    return ret;
}

char *str_dupt(char *p, ssize_t l)
{
    char *ret;

    if (l < 0) l = 0;
    ret = (char *)malloc(l + 1); 
    if (ret) {
	if (l > 0) memcpy(ret, p, l);
	ret[l] = 0;
    }
    return ret;
}

char *str_safedupt(char *p, ssize_t l)
{
    char *ret;

    if (l < 0) l = 0;
    ret = (char *)safe_malloc(l + 1); 
    if (l > 0) memcpy(ret, p, l);
    ret[l] = 0;
    return ret;
}

/* Returns true if p1,l1 equals p2,l2. Zero l1 or l2 allows p1 and p2 to be 0 */

int str_eq(char *p1, ssize_t l1, char *p2, ssize_t l2)
{
    return l1 == l2 && (!l2 || !memcmp(p1, p2, l2));
}


/* Returns true if (the first part of) p1,l1 equals p2,l2. Zero l2 returns 
 * always true and allows p1 and p2 to be 0. */

int str_feq(char *p1, ssize_t l1, char *p2, ssize_t l2)
{
    return l1 >= l2 && (!l2 || !memcmp(p1, p2, l2));
}


int str_cmp(char *p1, ssize_t l1, char *p2, ssize_t l2)
{
    int n;
    ssize_t rs;

    rs = l1 - l2;
    if (rs < 0) {
	n = memcmp(p1, p2, l1);
	if (n) return n;
	return -1;
    }
    if (rs > 0) {
	n = memcmp(p1, p2, l2);
	if (n) return n;
	return 1;
    }
    return memcmp(p1, p2, l1);
}


/*
 * vim:softtabstop=4:sw=4
 */
